from __future__ import absolute_import

import logging
from lxml import etree
import os
from random import randint

import six

from . import exceptions
from .metadata import MDWrap, MDRef, SubSection, AMDSec
from . import utils

LOGGER = logging.getLogger(__name__)


class FSEntry(object):
    """
    A class representing a filesystem entry - either a file or a directory.

    When passed to a :class:`metsrw.mets.METSDocument` instance, the tree of FSEntry objects will
    be used to construct the <fileSec> and <structMap> elements of a
    METS document.

    Unless otherwise specified, an FSEntry object is assumed to be a file;
    pass the `type` value as 'Directory' to specify that the object is
    instead a directory.

    An FSEntry object must be instantiated with a path as the first
    argument to the constructor, which represents its path on disk.

    An FSEntry object which is a Directory may have one or more children,
    representing files or directories contained within itself. Directory
    trees are designed for top-to-bottom traversal. Files cannot have
    children, and attempting to instantiate a file FSEntry object with
    children will raise a ValueError.

    Any FSEntry object may have one or more metadata entries associated
    with it; these can take the form of either references to other XML
    files on disk, which should be wrapped in MDRef objects, or
    wrapped copies of those XML files, which should be wrapped in
    MDWrap objects.

    :param str path: Path to the file on disk, as a bytestring. This will
        populate FLocat @xlink:href
    :param str label: Label in the structMap. If not provided, will be populated
        with the basename of path
    :param str use: Use for the fileGrp.  Items with identical uses will be
        grouped together.
    :param str type: Type of FSEntry this is. This will appear in the structMap.
    :param list children: List of :class:`FSEntry` that are direct children of
        this element in the structMap.  Only allowed if type is 'Directory'
    :param str file_uuid: UUID of this entry. Will be used to construct the
        FILEID used in the fileSec and structMap, and GROUPID.  Only required if
        type is 'Item'.
    :param FSEntry derived_from: FSEntry that this FSEntry is derived_from. This is used to set the GROUPID in the fileSec.
    :param str checksum: Value of the file's checksum. Required if checksumtype passed.
    :param str checksumtype: Type of the checksum. Must be one of :const:`FSEntry.ALLOWED_CHECKSUMS`.  Required if checksum passed.
    :raises ValueError: if children passed when type is not 'Directory'
    :raises ValueError: if only one of checksum or checksumtype passed
    :raises ValueError: if checksumtype is not in :const:`FSEntry.ALLOWED_CHECKSUMS`
    """

    ALLOWED_CHECKSUMS = ('Adler-32', 'CRC32', 'HAVAL', 'MD5', 'MNP', 'SHA-1', 'SHA-256', 'SHA-384', 'SHA-512', 'TIGER WHIRLPOOL')

    def __init__(self, path=None, label=None, use='original', type=u'Item', children=None, file_uuid=None, derived_from=None, checksum=None, checksumtype=None):
        # path can validly be any encoding; if this value needs
        # to be spliced later on, it's better to treat it as a
        # bytestring than as actually being encoded text.
        # TODO update this with six and bytes
        if path:
            path = str(path)
        self.path = path
        if label is None and path is not None:
            label = os.path.basename(path)
        self.label = label
        self.use = use
        self.type = six.text_type(type)
        self.parent = None
        self._children = []
        children = children or []
        for child in children:
            self.add_child(child)
        self.file_uuid = file_uuid
        self.derived_from = derived_from
        if bool(checksum) != bool(checksumtype):
            raise ValueError("Must provide both checksum and checksumtype, or neither. Provided values: %s and %s" % (checksum, checksumtype))
        if checksumtype and checksumtype not in self.ALLOWED_CHECKSUMS:
            raise ValueError(
                '%s must be one of %s' % (checksumtype, self.ALLOWED_CHECKSUMS))
        self.checksum = checksum
        self.checksumtype = checksumtype
        self.amdsecs = []
        self.dmdsecs = []

    def __str__(self):
        return '{s.type}: {s.path}'.format(s=self)

    def __repr__(self):
        return 'FSEntry(type={s.type!r}, path={s.path!r}, use={s.use!r}, label={s.label!r}, file_uuid={s.file_uuid!r}, checksum={s.checksum!r}, checksumtype={s.checksumtype!r})'.format(s=self)

    # PROPERTIES

    def _create_id(self, prefix):
        return prefix + '_' + str(randint(1, 999999))

    def file_id(self):
        """ Returns the fptr @FILEID if this is not a Directory. """
        if self.type.lower() == 'directory':
            return None
        if self.file_uuid is None:
            raise exceptions.MetsError('No FILEID: File %s does not have file_uuid set' % self.path)
        return utils.FILE_ID_PREFIX + self.file_uuid

    def group_id(self):
        """
        Returns the @GROUPID.

        If derived_from is set, returns that group_id.
        """
        if self.derived_from is not None:
            return self.derived_from.group_id()
        if self.file_uuid is None:
            return None
        return utils.GROUP_ID_PREFIX + self.file_uuid

    @property
    def admids(self):
        """ Returns a list of ADMIDs for this entry. """
        return [a.id_string() for a in self.amdsecs]

    @property
    def dmdids(self):
        """ Returns a list of DMDIDs for this entry. """
        return [d.id_string() for d in self.dmdsecs]

    @property
    def children(self):
        # Read-only
        return self._children

    # ADD ATTRIBUTES

    def _add_metadata_element(self, md, subsection, mdtype, mode='mdwrap', **kwargs):
        """
        :param md: Value to pass to the MDWrap/MDRef
        :param str subsection: Metadata tag to create.  See :const:`SubSection.ALLOWED_SUBSECTIONS`
        :param str mdtype: Value for mdWrap/mdRef @MDTYPE
        :param str mode: 'mdwrap' or 'mdref'
        :param str loctype: Required if mode is 'mdref'. LOCTYPE of a mdRef
        :param str label: Optional. Label of a mdRef
        :param str otherloctype: Optional. OTHERLOCTYPE of a mdRef.
        :param str othermdtype: Optional. OTHERMDTYPE of a mdWrap.

        """
        # HELP how handle multiple amdSecs?
        # When adding *MD which amdSec to add to?
        if mode.lower() == 'mdwrap':
            othermdtype = kwargs.get('othermdtype')
            mdsec = MDWrap(md, mdtype, othermdtype)
        elif mode.lower() == 'mdref':
            loctype = kwargs.get('loctype')
            label = kwargs.get('label')
            otherloctype = kwargs.get('otherloctype')
            mdsec = MDRef(md, mdtype, loctype, label, otherloctype)
        subsection = SubSection(subsection, mdsec)
        if subsection.subsection == 'dmdSec':
            self.dmdsecs.append(subsection)
        else:
            try:
                amdsec = self.amdsecs[0]
            except IndexError:
                amdsec = AMDSec()
                self.amdsecs.append(amdsec)
            amdsec.subsections.append(subsection)
        return subsection

    def add_techmd(self, md, mdtype, mode='mdwrap', **kwargs):
        return self._add_metadata_element(md, 'techMD', mdtype, mode, **kwargs)

    def add_digiprovmd(self, md, mdtype, mode='mdwrap', **kwargs):
        return self._add_metadata_element(md, 'digiprovMD', mdtype, mode, **kwargs)

    def add_rightsmd(self, md, mdtype, mode='mdwrap', **kwargs):
        return self._add_metadata_element(md, 'rightsMD', mdtype, mode, **kwargs)

    def add_dmdsec(self, md, mdtype, mode='mdwrap', **kwargs):
        return self._add_metadata_element(md, 'dmdSec', mdtype, mode, **kwargs)

    def add_premis_object(self, md, mode='mdwrap'):
        # TODO add extra args and create PREMIS object here
        return self.add_techmd(md, 'PREMIS:OBJECT', mode)

    def add_premis_event(self, md, mode='mdwrap'):
        # TODO add extra args and create PREMIS object here
        return self.add_digiprovmd(md, 'PREMIS:EVENT', mode)

    def add_premis_agent(self, md, mode='mdwrap'):
        # TODO add extra args and create PREMIS object here
        return self.add_digiprovmd(md, 'PREMIS:AGENT', mode)

    def add_premis_rights(self, md, mode='mdwrap'):
        # TODO add extra args and create PREMIS object here
        return self.add_rightsmd(md, 'PREMIS:RIGHTS', mode)

    def add_dublin_core(self, md, mode='mdwrap'):
        # TODO add extra args and create DC object here
        return self.add_dmdsec(md, 'DC', mode)

    def add_child(self, child):
        """
        Add a child FSEntry to this FSEntry.

        Only FSEntrys with a type of 'directory' can have children.

        This does not detect cyclic parent/child relationships, but that will cause problems.

        :param FSEntry child: FSEntry to add as a child
        :return: The newly added child
        :raises ValueError: If this FSEntry cannot have children.
        :raises ValueError: If the child and the parent are the same
        """
        if self.type.lower() != 'directory':
            raise ValueError("Only directory objects can have children")
        if child is self:
            raise ValueError('Cannot be a child of itself!')
        if child not in self._children:
            self._children.append(child)
        child.parent = self
        return child

    def remove_child(self, child):
        """
        Remove a child from this FSEntry

        If `child` is not actually a child of this entry, nothing happens.

        :param child: Child to remove
        """
        try:
            self._children.remove(child)
        except ValueError:  # Child may not be in list
            pass
        else:
            child.parent = None

    # SERIALIZE

    def serialize_filesec(self):
        """
        Return the file Element for this file, appropriate for use in a fileSec.

        If this is not an Item or has no use, return None.

        :return: fileSec element for this FSEntry
        """
        if self.type.lower() != 'item' or self.use is None:
            return None
        el = etree.Element(utils.lxmlns('mets') + 'file', ID=self.file_id())
        if self.group_id():
            el.attrib['GROUPID'] = self.group_id()
        if self.admids:
            el.set('ADMID', ' '.join(self.admids))
        if self.checksum and self.checksumtype:
            el.attrib['CHECKSUM'] = self.checksum
            el.attrib['CHECKSUMTYPE'] = self.checksumtype
        if self.path:
            flocat = etree.SubElement(el, utils.lxmlns('mets') + 'FLocat')
            # Setting manually so order is correct
            flocat.set(utils.lxmlns('xlink') + 'href', self.path)
            flocat.set('LOCTYPE', 'OTHER')
            flocat.set('OTHERLOCTYPE', 'SYSTEM')

        return el

    def serialize_structmap(self, recurse=True):
        """
        Return the div Element for this file, appropriate for use in a structMap.

        If this FSEntry represents a directory, its children will be recursively appended to itself.
        If this FSEntry represents a file, it will contain a <fptr> element.

        :param bool recurse: If true, serialize and apppend all children.  Otherwise, only serialize this element but not any children.
        :return: structMap element for this FSEntry
        """
        if not self.label:
            return None
        el = etree.Element(utils.lxmlns('mets') + 'div', TYPE=self.type, LABEL=self.label)
        if self.file_id():
            etree.SubElement(el, utils.lxmlns('mets') + 'fptr', FILEID=self.file_id())
        if self.dmdids:
            el.set('DMDID', ' '.join(self.dmdids))

        if recurse and self._children:
            for child in self._children:
                child_el = child.serialize_structmap(recurse=recurse)
                if child_el is not None:
                    el.append(child_el)

        return el



"""
Classes for metadata sections of the METS. Include amdSec, dmdSec, techMD, rightsMD, sourceMD, digiprovMD, mdRef and mdWrap.
"""
from __future__ import absolute_import

import logging
from lxml import etree
from random import randint

import six

from . import exceptions
from . import utils

LOGGER = logging.getLogger(__name__)


class AMDSec(object):
    """
    An object representing a section of administrative metadata in a
    document.

    This is ordinarily created by :class:`metsrw.mets.METSDocument` instances and does not
    have to be instantiated directly.

    :param str section_id: ID of the section. If not provided, will be generated from 'amdSec' and a random number.
    :param list subsections: List of :class:`SubSection` that are part of this amdSec
    :param Element tree: An lxml.Element that is an externally generated amdSec.  This will overwrite any automatic serialization.  If passed, section_id must also be passed.
    """
    tag = 'amdSec'

    def __init__(self, section_id=None, subsections=None, tree=None):
        if subsections is None:
            subsections = []
        self.subsections = subsections
        self._id = section_id
        self._tree = tree
        if tree is not None and not section_id:
            raise ValueError('If tree is provided, section_id must also be provided')

    def id_string(self, force_generate=False):
        """
        Returns the ID string for the amdSec.

        :param bool force_generate: If True, will generate a new ID from 'amdSec' and a random number.
        """
        # e.g., amdSec_1
        if force_generate or not self._id:
            self._id = self.tag + '_' + str(randint(1, 999999))
        return self._id

    @classmethod
    def parse(cls, root):
        """
        Create a new AMDSec by parsing root.

        :param root: Element or ElementTree to be parsed into an object.
        """
        if root.tag != utils.lxmlns('mets') + 'amdSec':
            raise exceptions.ParseError('AMDSec can only parse amdSec elements with METS namespace.')
        section_id = root.get('ID')
        subsections = []
        for child in root:
            subsection = SubSection.parse(child)
            subsections.append(subsection)
        return cls(section_id, subsections)

    def serialize(self, now=None):
        """
        Serialize this amdSec and all children to lxml Element and return it.

        :param str now: Default value for CREATED in children if none set
        :return: amdSec Element with all children
        """
        if self._tree is not None:
            return self._tree
        el = etree.Element(utils.lxmlns('mets') + self.tag, ID=self.id_string())
        self.subsections.sort()
        for child in self.subsections:
            el.append(child.serialize(now))
        return el


class SubSection(object):
    """
    An object representing a metadata subsection in a document.

    This is usually created automatically and does not have to be instantiated directly.

    :param str subsection: Tag name for the subsection to be created.  Should be
        one of 'techMD', 'rightsMD', 'sourceMD' or 'digiprovMD' if contained in an
        :class:`amdSec`, or 'dmdSec'.
    :param contents: The MDWrap or MDRef contained in this subsection.
    :type contents: :class:`MDWrap` or :class:`MDRef`
    :param str section_id: ID of the section. If not provided, will be generated from subsection tag and a random number.
    """
    ALLOWED_SUBSECTIONS = ('techMD', 'rightsMD', 'sourceMD', 'digiprovMD', 'dmdSec')

    def __init__(self, subsection, contents, section_id=None):
        if subsection not in self.ALLOWED_SUBSECTIONS:
            raise ValueError(
                '%s must be one of %s' % (subsection, self.ALLOWED_SUBSECTIONS))
        self.subsection = subsection
        self.contents = contents
        self._id = section_id
        self.status = None
        self.older = None
        self.newer = None
        self.created = None

    def __lt__(self, other):
        # Sort based on the subsection's order in ALLOWED_SUBSECTIONS
        # techMDs < rightsMD < sourceMD < digiprovMD < dmdSec
        return self.ALLOWED_SUBSECTIONS.index(self.subsection) < self.ALLOWED_SUBSECTIONS.index(other.subsection)

    def id_string(self, force_generate=False):
        """
        Returns the ID string for this SubSection.

        :param bool force_generate: If True, will generate a new ID from the subsection tag and a random number.
        """
        if force_generate or not self._id:
            self._id = self.subsection + '_' + str(randint(1, 999999))
        return self._id

    def get_status(self):
        """
        Returns the STATUS when serializing.

        Calculates based on the subsection type and if it's replacing anything.

        :returns: None or the STATUS string.
        """
        if self.status is not None:
            return self.status
        if self.subsection == 'dmdSec':
            if self.older is None:
                return 'original'
            else:
                return 'updated'
        if self.subsection in ('techMD', 'rightsMD',):
            # TODO how to handle ones where newer has been deleted?
            if self.newer is None:
                return 'current'
            else:
                return 'superseded'
        return None

    def replace_with(self, new_subsection):
        """
        Replace this SubSection with new_subsection.

        Replacing SubSection must be the same time.  That is, you can only
        replace a dmdSec with another dmdSec, or a rightsMD with a rightsMD etc.

        :param new_subsection: Updated version of this SubSection
        :type new_subsection: :class:`SubSection`
        """
        if self.subsection != new_subsection.subsection:
            raise exceptions.MetsError('Must replace a SubSection with one of the same type.')
        # TODO convert this to a DB so have bidirectonal foreign keys??
        self.newer = new_subsection
        new_subsection.older = self
        self.status = None

    @classmethod
    def parse(cls, root):
        """
        Create a new SubSection by parsing root.

        :param root: Element or ElementTree to be parsed into an object.
        :raises exceptions.ParseError: If root's tag is not in :const:`SubSection.ALLOWED_SUBSECTIONS`.
        :raises exceptions.ParseError: If the first child of root is not mdRef or mdWrap.
        """
        subsection = root.tag.replace(utils.lxmlns('mets'), '', 1)
        if subsection not in cls.ALLOWED_SUBSECTIONS:
            raise exceptions.ParseError('SubSection can only parse elements with tag in %s with METS namespace' % (cls.ALLOWED_SUBSECTIONS,))
        section_id = root.get('ID')
        created = root.get('CREATED', '')
        status = root.get('STATUS', '')
        child = root[0]
        if child.tag == utils.lxmlns('mets') + 'mdWrap':
            mdwrap = MDWrap.parse(child)
            obj = cls(subsection, mdwrap, section_id)
        elif child.tag == utils.lxmlns('mets') + 'mdRef':
            mdref = MDRef.parse(child)
            obj = cls(subsection, mdref, section_id)
        else:
            raise exceptions.ParseError('Child of %s must be mdWrap or mdRef' % subsection)
        obj.created = created
        obj.status = status
        return obj

    def serialize(self, now=None):
        """
        Serialize this SubSection and all children to lxml Element and return it.

        :param str now: Default value for CREATED if none set
        :return: dmdSec/techMD/rightsMD/sourceMD/digiprovMD Element with all children
        """
        created = self.created if self.created is not None else now
        el = etree.Element(utils.lxmlns('mets') + self.subsection, ID=self.id_string())
        if created:  # Don't add CREATED if none was parsed
            el.set('CREATED', created)
        status = self.get_status()
        if status:
            el.set('STATUS', status)
        if self.contents:
            el.append(self.contents.serialize())
        return el


class MDRef(object):
    """
    An object representing an external XML document, typically associated
    with an :class:`FSEntry` object.

    :param str target: Path to the external document. MDRef does not validate
        the existence of this target.
    :param str mdtype: The string representing the mdtype of XML document being
        enclosed. Examples include "PREMIS:OBJECT" and "PREMIS:EVENT".
    :param str label: Optional LABEL for the mdRef element
    :param str loctype: LOCTYPE of the mdRef.  Must be one of 'ARK', 'URN', 'URL', 'PURL', 'HANDLE', 'DOI' or 'OTHER'.
    :param str otherloctype: OTHERLOCTYPE of the mdRef. Should be provided if loctype is OTHER.
    """
    VALID_LOCTYPE = ('ARK', 'URN', 'URL', 'PURL', 'HANDLE', 'DOI', 'OTHER')

    def __init__(self, target, mdtype, loctype, label=None, otherloctype=None):
        self.target = target
        self.mdtype = mdtype
        self.loctype = loctype
        if loctype not in self.VALID_LOCTYPE:
            raise ValueError('loctype must be one of {}'.format(', '.join(self.VALID_LOCTYPE)))
        self.label = label
        self.otherloctype = otherloctype

    @classmethod
    def parse(cls, root):
        """
        Create a new MDWrap by parsing root.

        :param root: Element or ElementTree to be parsed into a MDWrap.
        """
        if root.tag != utils.lxmlns('mets') + 'mdRef':
            raise exceptions.ParseError('MDRef can only parse mdRef elements with METS namespace.')
        # Required attributes
        mdtype = root.get('MDTYPE')
        if not mdtype:
            raise exceptions.ParseError('mdRef must have a MDTYPE')
        target = root.get(utils.lxmlns('xlink') + 'href')
        if not target:
            raise exceptions.ParseError('mdRef must have an xlink:href.')
        loctype = root.get('LOCTYPE')
        if not loctype:
            raise exceptions.ParseError('mdRef must have a LOCTYPE')
        # Optional attributes
        label = root.get('LABEL')
        otherloctype = root.get('OTHERLOCTYPE')

        return cls(target, mdtype, loctype, label, otherloctype)

    def serialize(self):
        # If the source document is a METS document, the XPTR attribute of
        # this mdRef element should point to the IDs of each dmdSec element
        # in that document.
        XPTR = None
        try:
            target_doc = etree.parse(self.target)
            dmdsecs = [item.get('ID') for item in
                       target_doc.findall(utils.lxmlns('mets') + 'dmdSec')]
            XPTR = "xpointer(id(''))".format(' '.join(dmdsecs))
        except Exception:
            pass

        el = etree.Element(utils.lxmlns('mets') + 'mdRef')
        if self.label:
            el.attrib['LABEL'] = self.label
        if self.target:
            el.attrib[utils.lxmlns('xlink') + 'href'] = self.target
        el.attrib['MDTYPE'] = self.mdtype
        el.attrib['LOCTYPE'] = self.loctype
        if self.otherloctype:
            el.attrib['OTHERLOCTYPE'] = self.otherloctype
        if XPTR:
            el.attrib['XPTR'] = XPTR
        return el


class MDWrap(object):
    """
    An object representing an XML document enclosed in a METS document.
    The entirety of the XML document will be included; to reference an
    external document, use the :class:`MDRef` class.

    :param str document: A string copy of the document, and will be parsed into
        an ElementTree at the time of instantiation.
    :param str mdtype: The MDTYPE of XML document being enclosed. Examples
        include "PREMIS:OBJECT", "PREMIS:EVENT,", "DC" and "OTHER".
    :param str othermdtype: The OTHERMDTYPE of the XML document. Should be set if mdtype is "OTHER".
    """
    def __init__(self, document, mdtype, othermdtype=None):
        parser = etree.XMLParser(remove_blank_text=True)
        if isinstance(document, six.string_types):
            self.document = etree.fromstring(document, parser=parser)
        elif isinstance(document, etree._Element):
            self.document = document
        self.mdtype = mdtype
        self.othermdtype = othermdtype

    @classmethod
    def parse(cls, root):
        """
        Create a new MDWrap by parsing root.

        :param root: Element or ElementTree to be parsed into a MDWrap.
        :raises exceptions.ParseError: If mdWrap does not contain MDTYPE
        :raises exceptions.ParseError: If mdWrap or xmlData contain multiple children
        """
        if root.tag != utils.lxmlns('mets') + 'mdWrap':
            raise exceptions.ParseError('MDWrap can only parse mdWrap elements with METS namespace.')
        mdtype = root.get('MDTYPE')
        if not mdtype:
            raise exceptions.ParseError('mdWrap must have a MDTYPE')
        othermdtype = root.get('OTHERMDTYPE')
        document = root.xpath('mets:xmlData/*', namespaces=utils.NAMESPACES)
        if len(document) != 1:
            raise exceptions.ParseError('mdWrap and xmlData can only have one child')
        document = document[0]
        return cls(document, mdtype, othermdtype)

    def serialize(self):
        el = etree.Element(utils.lxmlns('mets') + 'mdWrap', MDTYPE=self.mdtype)
        if self.othermdtype:
            el.attrib['OTHERMDTYPE'] = self.othermdtype
        xmldata = etree.SubElement(el, utils.lxmlns('mets') + 'xmlData')
        xmldata.append(self.document)

        return el



from __future__ import absolute_import

from datetime import datetime
import logging
from lxml import etree
import sys

# This package
from . import exceptions
from . import fsentry
from . import metadata
from . import utils

LOGGER = logging.getLogger(__name__)


class METSDocument(object):
    def __init__(self):
        # Stores the ElementTree if this was parsed from an existing file
        self.tree = None
        # Only root-level elements are stored, since the rest
        # can be inferred via their #children attribute
        self.createdate = None
        self._root_elements = []
        self._all_files = None
        self.dmdsecs = []
        self.amdsecs = []

    # FSENTRYS

    def _collect_all_files(self, files=None):
        """
        Collect all FSEntrys into a set, including all descendants.

        :param list files: List of :class:`FSEntry` to traverse.
        :returns: Set of FSEntry
        """
        if files is None:
            files = self._root_elements
        collected = set()
        for entry in files:
            collected.add(entry)
            collected.update(self._collect_all_files(entry.children))
        return collected

    def all_files(self):
        """
        Return a set of all FSEntrys in this METS document.

        :returns: Set containing all :class:`FSEntry` in this METS document,
            including descendants of ones explicitly added.
        """
        # FIXME cache this. Should not break when add_child is called on an element already in the document.
        return self._collect_all_files(self._root_elements)

    def get_file(self, **kwargs):
        """
        Return the FSEntry that matches parameters.

        :param str file_uuid: UUID of the target FSEntry.
        :param str label: structMap LABEL of the target FSEntry.
        :param str type: structMap TYPE of the target FSEntry.
        :returns: :class:`FSEntry` that matches parameters, or None.
        """
        # TODO put this in a sqlite DB so it can be queried efficiently
        # TODO handle multiple matches (with DB?)
        # TODO check that kwargs are actual attrs
        for entry in self.all_files():
            if all(value == getattr(entry, key) for key, value in kwargs.items()):
                return entry
        return None

    def append_file(self, fs_entry):
        """
        Adds an FSEntry object to this METS document's tree. Any of the
        represented object's children will also be added to the document.

        A given FSEntry object can only be included in a document once,
        and any attempt to add an object the second time will be ignored.

        :param FSEntry fs_entry: FSEntry to add to the METS document
        """

        if fs_entry in self._root_elements:
            return
        self._root_elements.append(fs_entry)
        # Reset file lists so they get regenerated with the new files(s)
        self._all_files = None

    def remove_entry(self, fs_entry):
        """
        Removes an FSEntry object from this METS document.

        Any children of this FSEntry will also be removed. This will be removed as a child of it's parent, if any.

        :param FSEntry fs_entry: FSEntry to remove from the METS
        """
        try:
            self._root_elements.remove(fs_entry)
        except ValueError:  # fs_entry may not be in the root elements
            pass
        if fs_entry.parent:
            fs_entry.parent.remove_child(fs_entry)
        # Reset file lists so they get regenerated without the removed file(s)
        self._all_files = None


    # SERIALIZE

    def _document_root(self, fully_qualified=True):
        """
        Return the mets Element for the document root.
        """
        nsmap = {
            'xsi': utils.NAMESPACES['xsi'],
            'xlink': utils.NAMESPACES['xlink']
        }
        if fully_qualified:
            nsmap['mets'] = utils.NAMESPACES['mets']
        else:
            nsmap[None] = utils.NAMESPACES['mets']
        attrib = {
            '{}schemaLocation'.format(utils.lxmlns('xsi')):
            utils.SCHEMA_LOCATIONS
        }
        return etree.Element(utils.lxmlns('mets') + 'mets', nsmap=nsmap, attrib=attrib)

    def _mets_header(self, now):
        """
        Return the metsHdr Element.
        """
        if self.createdate is None:
            e = etree.Element(utils.lxmlns('mets') + 'metsHdr', CREATEDATE=now)
        else:
            e = etree.Element(utils.lxmlns('mets') + 'metsHdr',
                CREATEDATE=self.createdate, LASTMODDATE=now)
        return e

    def _collect_mdsec_elements(self, files):
        """
        Return all dmdSec and amdSec classes associated with the files.

        Returns all dmdSecs, then all amdSecs, so they only need to be
        serialized before being appended to the METS document.

        :param List files: List of :class:`FSEntry` to collect MDSecs for.
        :returns: List of AMDSecs and SubSections
        """
        dmdsecs = []
        amdsecs = []
        for f in files:
            for d in f.dmdsecs:
                dmdsecs.append(d)
            for a in f.amdsecs:
                amdsecs.append(a)

        dmdsecs.sort(key=lambda x: x.id_string())
        amdsecs.sort(key=lambda x: x.id_string())
        return dmdsecs + amdsecs

    def _structmap(self):
        """
        Returns structMap element for all files.
        """
        structmap = etree.Element(utils.lxmlns('mets') + 'structMap',
                                  TYPE='physical',
                                  # TODO Add ability for multiple structMaps
                                  ID='structMap_1',
                                  # TODO don't hardcode this
                                  LABEL='Archivematica default')
        for item in self._root_elements:
            child = item.serialize_structmap(recurse=True)
            if child is not None:
                structmap.append(child)

        return structmap

    def _filesec(self, files=None):
        """
        Returns fileSec Element containing all files grouped by use.
        """
        if files is None:
            files = self.all_files()

        filesec = etree.Element(utils.lxmlns('mets') + 'fileSec')
        filegrps = {}
        for file_ in files:
            if file_.type.lower() != 'item':
                continue
            # Get fileGrp, or create if not exist
            filegrp = filegrps.get(file_.use)
            if filegrp is None:
                filegrp = etree.SubElement(filesec, utils.lxmlns('mets') + 'fileGrp', USE=file_.use)
                filegrps[file_.use] = filegrp

            file_el = file_.serialize_filesec()
            if file_el is not None:
                filegrp.append(file_el)

        return filesec

    def serialize(self, fully_qualified=True):
        """
        Returns this document serialized to an xml Element.

        :return: Element for this document
        """
        now = datetime.utcnow().replace(microsecond=0).isoformat('T')
        files = self.all_files()
        mdsecs = self._collect_mdsec_elements(files)
        root = self._document_root(fully_qualified=fully_qualified)
        root.append(self._mets_header(now=now))
        for section in mdsecs:
            root.append(section.serialize(now=now))
        root.append(self._filesec(files))
        root.append(self._structmap())

        return root

    def tostring(self, fully_qualified=True, pretty_print=True):
        """
        Serialize and return a string of this METS document.

        To write to file, see :meth:`write`

        :return: String of this document
        """
        root = self.serialize(fully_qualified=fully_qualified)
        return etree.tostring(root, pretty_print=pretty_print, xml_declaration=True)

    def write(self, filepath, fully_qualified=True, pretty_print=False):
        """
        Serialize and write this METS document to `filepath`.

        :param str filepath: Path to write the METS document to
        """
        root = self.serialize(fully_qualified=fully_qualified)
        tree = root.getroottree()
        tree.write(filepath, xml_declaration=True, pretty_print=pretty_print)

    # PARSE HELPERS

    def _parse_tree_structmap(self, tree, parent_elem):
        """
        Recursively parse all the children of parent_elem, including amdSecs and dmdSecs.
        """
        siblings = []
        for elem in parent_elem:  # Iterates over children of parent_elem
            # Only handle div's, not fptrs
            if elem.tag != utils.lxmlns('mets') + 'div':
                continue
            entry_type = elem.get('TYPE')
            label = elem.get('LABEL')
            fptr = elem.find('mets:fptr', namespaces=utils.NAMESPACES)
            file_uuid = derived_from = use = path = amdids = checksum = checksumtype = None
            if fptr is not None:
                file_id = fptr.get('FILEID')
                file_elem = tree.find('mets:fileSec//mets:file[@ID="' + file_id + '"]', namespaces=utils.NAMESPACES)
                if file_elem is None:
                    raise exceptions.ParseError('%s exists in structMap but not fileSec' % file_id)
                file_uuid = file_id.replace(utils.FILE_ID_PREFIX, '', 1)
                group_uuid = file_elem.get('GROUPID', '').replace(utils.GROUP_ID_PREFIX, '', 1)
                if group_uuid != file_uuid:
                    derived_from = group_uuid  # Use group_uuid as placeholder
                use = file_elem.getparent().get('USE')
                path = file_elem.find('mets:FLocat', namespaces=utils.NAMESPACES).get(utils.lxmlns('xlink') + 'href')
                amdids = file_elem.get('ADMID')
                checksum = file_elem.get('CHECKSUM')
                checksumtype = file_elem.get('CHECKSUMTYPE')

            # Recursively generate children
            children = self._parse_tree_structmap(tree, elem)

            # Create FSEntry
            fs_entry = fsentry.FSEntry(path=path, label=label, use=use, type=entry_type, children=children, file_uuid=file_uuid, derived_from=derived_from, checksum=checksum, checksumtype=checksumtype)

            # Add DMDSecs
            dmdids = elem.get('DMDID')
            if dmdids:
                dmdids = dmdids.split()
                for dmdid in dmdids:
                    dmdsec_elem = tree.find('mets:dmdSec[@ID="' + dmdid + '"]', namespaces=utils.NAMESPACES)
                    dmdsec = metadata.SubSection.parse(dmdsec_elem)
                    fs_entry.dmdsecs.append(dmdsec)
                # Create older/newer relationships
                fs_entry.dmdsecs.sort(key=lambda x: x.created)
                for prev_dmdsec, dmdsec in zip(fs_entry.dmdsecs, fs_entry.dmdsecs[1:]):
                    if dmdsec.status == 'updated':
                        prev_dmdsec.replace_with(dmdsec)

            # Add AMDSecs
            if amdids:
                amdids = amdids.split()
                for amdid in amdids:
                    amdsec_elem = tree.find('mets:amdSec[@ID="' + amdid + '"]', namespaces=utils.NAMESPACES)
                    amdsec = metadata.AMDSec.parse(amdsec_elem)
                    fs_entry.amdsecs.append(amdsec)

            siblings.append(fs_entry)
        return siblings

    def _parse_tree(self, tree=None):
        if tree is None:
            tree = self.tree
        # self._validate()
        # Check CREATEDATE < now
        createdate = self.tree.find('mets:metsHdr', namespaces=utils.NAMESPACES).get('CREATEDATE')
        now = datetime.utcnow().isoformat('T')
        if createdate and createdate > now:
            raise exceptions.ParseError('CREATEDATE more recent than now (%s)' % now)
        self.createdate = createdate

        # Parse structMap
        structMap = tree.find('mets:structMap[@TYPE="physical"]', namespaces=utils.NAMESPACES)
        if structMap is None:
            raise exceptions.ParseError('No physical structMap found.')
        self._root_elements = self._parse_tree_structmap(tree, structMap)

        # Associated derived files
        for entry in self.all_files():
            entry.derived_from = self.get_file(file_uuid=entry.derived_from, type='Item')

    def _validate(self):
        raise NotImplementedError()

    def _fromfile(self, path):
        """
        Accept a filepath pointing to a valid METS document and parses it.

        :param str path: Path to a METS document.
        """
        parser = etree.XMLParser(remove_blank_text=True)
        self.tree = etree.parse(path, parser=parser)
        self._parse_tree(self.tree)

    @classmethod
    def fromfile(klass, path):
        """ Creates a METS by parsing a file. """
        i = klass()
        i._fromfile(path)
        return i

    def _fromstring(self, string):
        """
        Accept a string containing valid METS xml and parses it.

        :param str string: String containing a METS document.
        """
        parser = etree.XMLParser(remove_blank_text=True)
        root = etree.fromstring(string, parser)
        self.tree = root.getroottree()
        self._parse_tree(self.tree)

    @classmethod
    def fromstring(klass, string):
        """ Create a METS by parsing a string. """
        i = klass()
        i._fromstring(string)
        return i

    def _fromtree(self, tree):
        """
        Accept an ElementTree or Element and parses it.

        :param ElementTree tree: ElementTree to build a METS document from.
        """
        self.tree = tree
        self._parse_tree(self.tree)

    @classmethod
    def fromtree(klass, tree):
        """ Create a METS from an ElementTree or Element. """
        i = klass()
        i._fromtree(tree)
        return i

if __name__ == '__main__':
    mw = METSDocument()
    mw.fromfile(sys.argv[1])
    mw.write(sys.argv[2], fully_qualified=True, pretty_print=True)



# LXML HELPERS

NAMESPACES = {
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "mets": "http://www.loc.gov/METS/",
    "premis": "info:lc/xmlns/premis-v2",
    "dcterms": "http://purl.org/dc/terms/",
    "fits": "http://hul.harvard.edu/ois/xml/ns/fits/fits_output",
    "xlink": "http://www.w3.org/1999/xlink",
    "dc": "http://purl.org/dc/elements/1.1/"
}

SCHEMA_LOCATIONS = "http://www.loc.gov/METS/ " + \
                   "http://www.loc.gov/standards/mets/version111/mets.xsd"


def lxmlns(arg):
    """ Return XPath-usable namespace. """
    return '{' + NAMESPACES[arg] + '}'


# CONSTANTS

FILE_ID_PREFIX = 'file-'
GROUP_ID_PREFIX = 'Group-'



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='metsrw',
    version='0.1.1',

    description='Library for dealing with METS files.',
    long_description=long_description,

    url='https://github.com/artefactual-labs/mets-reader-writer/',

    author='Artefactual',
    author_email='info@artefactual.com',

    license='AGPL',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],

    keywords='mets',

    packages=find_packages(exclude=['docs', 'fixtures', 'requirements', 'tests*']),

    install_requires=['lxml', 'six'],
)



from __future__ import absolute_import

import logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

from .exceptions import *
from .fsentry import *
from .metadata import *
from .mets import *
from .utils import *



"""
Exceptions for metsrw.

All exceptions generated by this library will descend from MetsError.
"""
class MetsError(Exception):
    """ Base Exception for this module. """
    pass


class ParseError(MetsError):
    """ Error parsing a METS file. """
    pass
